/* 
 * glDino - 3D motor to render three-dimentional objects
 *
 * Mesa 3D graphics library 2.6 version
 * Sun Solaris 2.5 or higher
 * Linux 2.0.x or higher
 *
 * Author: Navarro Mikael (navarro@essi.fr)
 *
 */

/*
 * ALL RIGHTS RESERVED 
 * Permission to use, copy, modify, and distribute this software for 
 * any purpose and without fee is hereby granted, provided that the above
 * copyright notice appear in all copies and that both the copyright notice
 * and this permission notice appear in supporting documentation, and that 
 * the name of OpenGL Inc. not be used in advertising or publicity pertaining
 * to distribution of the software without specific, written prior permission.
 *
 * THE MATERIAL EMBODIED ON THIS SOFTWARE IS PROVIDED TO YOU "AS-IS"
 * AND WITHOUT WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR OTHERWISE,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF MERCHANTABILITY OR
 * FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL SILICON
 * GRAPHICS, INC.  BE LIABLE TO YOU OR ANYONE ELSE FOR ANY DIRECT,
 * SPECIAL, INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY
 * KIND, OR ANY DAMAGES WHATSOEVER, INCLUDING WITHOUT LIMITATION,
 * LOSS OF PROFIT, LOSS OF USE, SAVINGS OR REVENUE, OR THE CLAIMS OF
 * THIRD PARTIES, WHETHER OR NOT SILICON GRAPHICS, INC.  HAS BEEN
 * ADVISED OF THE POSSIBILITY OF SUCH LOSS, HOWEVER CAUSED AND ON
 * ANY THEORY OF LIABILITY, ARISING OUT OF OR IN CONNECTION WITH THE
 * POSSESSION, USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/* Compile: gcc -ansi -O2 -o glDino glDino.c dino.c 
   -L/usr/X11R6/lib -lglut -lGLU -lGL -lXmu -lXext -lX11 -lm */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>       /* for cos(), sin(), and sqrt() */
#include <GL/glu.h>     /* OpenGL Utility Library */
#include <GL/glut.h>    /* OpenGL Utility Toolkit header */

#include "dino.h"   /* draw Dinosaur */
#include "floor.h"

/* Some <math.h> files do not define M_PI... */
#ifndef M_PI
#define M_PI 3.14159265359
#endif

/* Some usefull macros */
#define MIN_VALUE(a,b) (((a) < (b)) ? a : b)
#define MAX_VALUE(a,b) (((a) > (b)) ? a : b)
#define RAD2DEG(x) (x * 180 / M_PI)
#define DEG2RAD(x) (x * M_PI / 180)

/* Some predefined colors */
GLfloat black[] = {0.0, 0.0, 0.0, 1.0};
GLfloat gray[] = {0.7, 0.7, 0.7, 1.0};
GLfloat grey[] = {0.3, 0.3, 0.3, 1.0};
GLfloat white[] = {1.0, 1.0, 1.0, 1.0};
GLfloat red[] = {1.0, 0.0, 0.0, 1.0};
GLfloat green[] = {0.0, 1.0, 0.0, 1.0};
GLfloat blue[] = {0.0, 0.0, 1.0, 1.0};
GLfloat yellow[] = {1.0, 1.0, 0.0, 1.0};
GLfloat magenta[] = {1.0, 0.0, 1.0, 1.0};
GLfloat cyan[] = {0.0, 1.0, 1.0, 1.0};

/* Lights properties */
GLfloat light_position0[] = {-1.0, 1.0, 1.0, 0.0}; /* sun */
GLfloat light_ambient0[] = {0.0, 0.0, 0.0, 1.0};
GLfloat light_diffuse0[] = {1.0, 1.0, 1.0, 1.0};
GLfloat light_specular0[] = {1.0, 1.0, 1.0, 2.0};

/* Material properties */
GLfloat mat_diffuse[] = {1.0, 1.0, 1.0, 0.5};
GLfloat mat_specular[] = {10.0, 1.0, 1.0, 0.5};
GLfloat mat_shininess[] = {10.0};

/* Variable controlling various rendering modes */
static int animation=0;
static int lightanimation=0;
static int scaling=0;
static int renderDinosaur=1;
static int reportSpeed=0;
static int useDepth=1;
static int directionalLight=0;
static int lightSwitch=1;
static int fullScreen=0;

enum { 
  M_NONE, M_MOTION, M_ANIMATION, M_LIGHT, M_DINOSAUR, M_QUIT,
  M_POSITIONAL, M_DIRECTIONAL, M_PERFORMANCE, M_FULL_SCREEN
};

/* User-controled variables */
static GLfloat tx=0.0, ty=0.0, tz=0.0;
static GLfloat rx=0.0, ry=0.0, rz=1.0;
static GLfloat anglex=0.0, angley=-50.0, anglez=0.0; /* in degrees */
static GLfloat sx=1.0, sy=1.0, sz=1.0, scalefactor=1.0;
static float lightangle=135.0, lightheight=20;
static float jump=0.0;
static int moving;
static int startx, starty, endx, endy;
static int lightmoving=0;
static int lightstartx, lightstarty;
static int start, end;

/* Windows properties */
GLint width=512, height=512, posx=80, posy=20;

/* Forward declarations */
static void draw(void);
static void display(void);
static void init();
static void menu(int value);
static void mouse(int button, int state, int x, int y);
static void motion(int x, int y);
static void keys(unsigned char c, int x, int y);
static void specialkeys(int c, int x, int y);
static void idle(void);
static void timer(void);
static void reshape(GLsizei w, GLsizei h);
static void entry(int state);
static void visibility(int state);

static void draw(void) {
  drawDinosaur();
}

static void display(void) {
  if(reportSpeed) {
    start = glutGet(GLUT_ELAPSED_TIME);
  }

  glClear(GL_COLOR_BUFFER_BIT | (useDepth ? GL_DEPTH_BUFFER_BIT : 0));

  /* Adjust light position */
  light_position0[0] = 12*cos(lightangle);
  light_position0[1] = lightheight;
  light_position0[2] = 12*sin(lightangle);
  if(directionalLight) {
    light_position0[3] = 0.0;
  }
  else {
    light_position0[3] = 1.0;
  }

  glPushMatrix();
  /* Adjust object position */
  glRotatef(anglex, 1.0, 0.0, 0.0);
  glRotatef(angley, 0.0, 1.0, 0.0);
  /* Draw stuff here */

  /* Set light position */
  glLightfv(GL_LIGHT0, GL_POSITION, light_position0);

  /* Draw reflected dinosaur */
  glPushMatrix();
  glScalef(1.0, -1.0, 1.0); /* reflect across y=0 plane */
  glLightfv(GL_LIGHT0, GL_POSITION, light_position0);
  /* XXX Ugh, unfortunately the back face culling reverses when we reflect 
     the dinosaur. Easy solution is just disable back face culling for
     rendering the reflection. (be carefull that object material properties 
     are defined GLFRONT_AND_BACK). Also the normals for lighting get screwed
     up by the scale; enabled normalize. */ 
  if(renderDinosaur) {
    glEnable(GL_NORMALIZE);
    glCullFace(GL_FRONT);
    draw();
    glDisable(GL_NORMALIZE);
    glCullFace(GL_BACK);
  }
  glPopMatrix();

  /* Restore light position */
  glLightfv(GL_LIGHT0, GL_POSITION, light_position0);

  /* Draw top of floor (y=0) */
  glEnable(GL_BLEND);
  glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
  glColor4f(0.7, 0.7, 0.7, 0.4); /* 40% dark gray floor color */
  drawFloor();
  glDisable(GL_BLEND);
  /* Draw bottom of floor */
  glFrontFace(GL_CW); /* switch face orientation */
  glColor4f(0.1, 0.1, 0.7, 1.0);
  drawFloor();
  glFrontFace(GL_CCW);

  /* draw dinosaur */
  if(renderDinosaur) {
    draw();
  }

  glPushMatrix();
  glDisable(GL_LIGHTING);
  glColor3f(1.0, 1.0, 0.0);
  if (directionalLight) {
    /* Draw an arrowhead */
    glDisable(GL_CULL_FACE);
    glTranslatef(light_position0[0], light_position0[1], light_position0[2]);
    glRotatef(lightangle * -180.0 / M_PI, 0, 1, 0);
    glRotatef(atan(lightheight/12) * 180.0 / M_PI, 0, 0, 1);
    glBegin(GL_TRIANGLE_FAN);
    glVertex3f(0, 0, 0);
    glVertex3f(2, 1, 1);
    glVertex3f(2, -1, 1);
    glVertex3f(2, -1, -1);
    glVertex3f(2, 1, -1);
    glVertex3f(2, 1, 1);
    glEnd();
    /* Draw a white line from light direction */
    glColor3f(1.0, 1.0, 1.0);
    glBegin(GL_LINES);
    glVertex3f(0, 0, 0);
    glVertex3f(5, 0, 0);
    glEnd();
    glEnable(GL_CULL_FACE);
  } else {
    /* Draw a yellow ball at the light source */
    glTranslatef(light_position0[0], light_position0[1], light_position0[2]);
    glutSolidSphere(1.0, 10, 10);
  }
  glEnable(GL_LIGHTING);
  glPopMatrix();

  glPopMatrix();

  if(reportSpeed) {
    glFinish();
    end = glutGet(GLUT_ELAPSED_TIME);
    printf("Speed %.3g frames/s (%d ms)\n", 1000.0/(end-start), end-start);
  }

  glutSwapBuffers();
}

static void init() {
  /* Enable OpenGL lights and set materials */
  glLightfv(GL_LIGHT0, GL_POSITION, light_position0);
  glLightfv(GL_LIGHT0, GL_AMBIENT, light_ambient0);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, light_diffuse0);
  glLightfv(GL_LIGHT0, GL_SPECULAR, light_specular0);
  glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
  glLightf(GL_LIGHT0, GL_CONSTANT_ATTENUATION, 0.1);
  glLightf(GL_LIGHT0, GL_LINEAR_ATTENUATION, 0.05);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0);

  /* Some general */
  glEnable(GL_NORMALIZE);
  glCullFace(GL_BACK);
  glEnable(GL_CULL_FACE);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_FASTEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST);
  glHint(GL_FOG_HINT, GL_FASTEST);
  
  /* Use depth buffer */
  glDepthFunc(GL_LEQUAL);
  glEnable(GL_DEPTH_TEST);
}

static void menu(int value) {
  switch(value) {
  case M_NONE:
    return;
  case M_QUIT:
    exit(0);
  case M_FULL_SCREEN:
    fullScreen = 1 - fullScreen;
    if(fullScreen) {
      width = glutGet(GLUT_WINDOW_WIDTH);
      height = glutGet(GLUT_WINDOW_HEIGHT);
      posx = glutGet(GLUT_WINDOW_X);
      posy = glutGet(GLUT_WINDOW_Y);
      glutFullScreen();
    }
    else {
      glutReshapeWindow(width, height);
      glutPositionWindow(posx, posy);
    }
    break;
  case M_MOTION:
    animation = 1 - animation;
    break;
  case M_ANIMATION:
    lightanimation = 1 - lightanimation;
    break;
  case M_LIGHT:
    lightSwitch = 1 - lightSwitch;
    if(lightSwitch) {
      glEnable(GL_LIGHT0);
    }
    else {
      glDisable(GL_LIGHT0);
    }
    break;
  case M_DINOSAUR:
    renderDinosaur = 1 - renderDinosaur;
    break;
  case M_POSITIONAL:
    directionalLight = 0;
    break;
  case M_DIRECTIONAL:
    directionalLight = 1;
    break;
  case M_PERFORMANCE:
    reportSpeed = 1 - reportSpeed;
    break;
  }
  glutPostRedisplay();
}

/* ARGSUSED2 */
static void mouse(int button, int state, int x, int y) {
  if(button == GLUT_LEFT_BUTTON && state == GLUT_DOWN) {
    glutSetCursor(GLUT_CURSOR_CYCLE);
    moving = 1;
    startx = x;
    starty = y;
    if(glutGetModifiers() & GLUT_ACTIVE_SHIFT) {
      glutSetCursor(GLUT_CURSOR_UP_DOWN);
      scaling = 1;
    } else {
      scaling = 0;
    }
  }
  if(button == GLUT_LEFT_BUTTON && state == GLUT_UP) {
    glutSetCursor(GLUT_CURSOR_INHERIT);
    moving = 0;
    scaling = 0;
  }
  if(button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) {
    glutSetCursor(GLUT_CURSOR_CROSSHAIR);
    lightmoving = 1;
    lightstartx = x;
    lightstarty = y;
  }
  if(button == GLUT_MIDDLE_BUTTON && state == GLUT_UP) {
    glutSetCursor(GLUT_CURSOR_INHERIT);
    lightmoving = 0;
  }
}

/* ARGSUSED1 */
static void motion(int x, int y) {
  if(lightmoving) {
    lightangle += (lightstartx - x)/40.0;
    lightheight += (lightstarty - y)/20.0;
    lightstartx = x;
    lightstarty = y;
    glutPostRedisplay();
  }
  if(scaling) {
    scalefactor = scalefactor * (1.0f + (((float) (starty - y)) / height));
    sx = sy = sz = scalefactor;
    startx = x;
    starty = y;
    glutPostRedisplay();
    return;
  }
  if(moving) {
    angley = angley + (x- startx);
    anglex = anglex + (y - starty);
    startx = x;
    starty = y;
    glutPostRedisplay();
  }
}

/* Press any key to redraw; good when motion stopped and
   performance reporting on */
/* ARGSUSED */
static void keys(unsigned char c, int x, int y) {
  if(c == 27) {
    exit(0);  /* IRIS GLism, Escape quits */
  }
  if(c == 'q' || c == 'Q') {
    exit(0);
  }
  glutPostRedisplay();
}

static void specialkeys(int c, int x, int y) {
  if(c == GLUT_KEY_UP) {
  }
  if(c == GLUT_KEY_DOWN) {
  }
  if(c == GLUT_KEY_LEFT) {
  }
  if(c == GLUT_KEY_RIGHT) {
  }
  glutPostRedisplay();
}

/* Advance time varying state when idle callback registered */
static void idle(void) {
  static float time = 0.0;

  time = glutGet(GLUT_ELAPSED_TIME)/500.0;
  if(animation) {
    jump = 4.0*fabs(sin(time)*0.5);
  }
  if(lightanimation) {
    lightangle += 0.03;
  }
  glutPostRedisplay();
}

static void timer(void) {
}

void reshape(GLsizei w, GLsizei h) {
  /* Setup the view of the object */
  glViewport(0, 0, w, h);
  
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, (GLfloat)w/(GLfloat)h, 1.0, 100.0);

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(0.0, 20.0, 50.0,
	    0.0, 10.0, 0.0,
	    0.0, 1.0, 0.0);
}

static void entry(int state) {
  if(state == GLUT_ENTERED && (animation || lightanimation)) {
    glutIdleFunc(idle);
  }
  else if(state == GLUT_LEFT && (animation || lightanimation)) {
    glutIdleFunc(NULL);
  }
}

static void visibility(int state) {
  if(state == GLUT_NOT_VISIBLE && (animation || lightanimation)) {
    glutIdleFunc(NULL);
  }
  else if(state == GLUT_VISIBLE && (animation || lightanimation)) {
    glutIdleFunc(idle);
  }
}

int main(int argc, char **argv) {
  int i;

  glutInitWindowSize(width, height);
  glutInitWindowPosition(posx, posy);

  glutInit(&argc, argv);
  for (i=1; i<argc; i++) {
    if (!strcmp("-h", argv[i]) || !strcmp("--help", argv[i])) {
      printf("glDino  :-)\n");
      printf("Usage: %s --help\n", argv[0]);
      printf("Options:\n");
      printf("  --help   display this message\n");
      printf("Keys:\n");
      printf("  <ESC> quit\n");
      exit(0);
    }
  }

  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH
		      | GLUT_STENCIL);
    
#if 0
  /* In GLUT 4.0, you'll be able to do this an be sure to
       get 2 bits of stencil if the machine has it for you */
  glutInitDisplayString("samples stencil>=2 rgb double depth");
#endif

  glutCreateWindow(argv[0]);
    
  if (glutGet(GLUT_WINDOW_STENCIL_SIZE) <= 1) {
    printf("glDino: Sorry, I need at least 2 bits of stencil.\n");
    exit(1);
  }
    
  /* Register GLUT callbacks */
  glutDisplayFunc(display);
  init();

  glutMouseFunc(mouse);
  glutMotionFunc(motion);

  glutKeyboardFunc(keys);
  glutSpecialFunc(specialkeys);

  glutIdleFunc(idle);
  /*glutTimerFunc(1000, timer, 0);*/

  glutReshapeFunc(reshape);
  glutEntryFunc(entry);
  glutVisibilityFunc(visibility);

  glutCreateMenu(menu);
  glutAddMenuEntry("Toggle animation", M_ANIMATION);
  glutAddMenuEntry("Toggle motion", M_MOTION);
  glutAddMenuEntry("Toggle full screen", M_FULL_SCREEN);
  glutAddMenuEntry("-----------------------", M_NONE);
  glutAddMenuEntry("Toggle light", M_LIGHT);
  glutAddMenuEntry("Toggle dinosaur", M_DINOSAUR);
  glutAddMenuEntry("----------------------", M_NONE);
  glutAddMenuEntry("Positional light", M_POSITIONAL);
  glutAddMenuEntry("Directional light", M_DIRECTIONAL);
  glutAddMenuEntry("-----------------------", M_NONE);
  glutAddMenuEntry("Toggle performance", M_PERFORMANCE);
  glutAddMenuEntry("-----------------------", M_NONE);
  glutAddMenuEntry("Quit", M_QUIT);
  glutAttachMenu(GLUT_RIGHT_BUTTON);

  glutMainLoop();
  return 0; /* ANSI C requires main to return int. */
}

